/* Reimplemented NoSQL column operator
 * Copyright (c) 2003 Micah Stetson <micah@cnm-vra.com>
 * 
 * Permission is hereby granted, free of charge, to any
 * person obtaining a copy of this software and associated
 * documentation files (the "Software"), to deal in the
 * Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute,
 * sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice
 * shall be included in all copies or substantial portions
 * of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF
 * ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 * 
 */
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <errno.h>

#ifndef HELPFILE
#define HELPFILE "/usr/local/nosql/help/column.txt"
#endif

/* Argument processing and getfields() taken from Plan 9's u9fs.
 * These routines were modified to be ASCII only, and the argument
 * routines were extended to do something half sensible with
 * --gnu-silly-options.
 *
 * The authors of this software are Russ Cox, Sean Dorward, Eric Grosse, et al.
 * Copyright (c) 2002 by Lucent Technologies.
 * Permission to use, copy, modify, and distribute this software for any
 * purpose without fee is hereby granted, provided that this entire notice
 * is included in all copies of any software which is or includes a copy
 * or modification of this software and in all copies of the supporting
 * documentation for such software.
 * THIS SOFTWARE IS BEING PROVIDED "AS IS", WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTY.  IN PARTICULAR, NEITHER THE AUTHORS NOR LUCENT MAKE ANY
 * REPRESENTATION OR WARRANTY OF ANY KIND CONCERNING THE MERCHANTABILITY
 * OF THIS SOFTWARE OR ITS FITNESS FOR ANY PARTICULAR PURPOSE.
 */
char *argv0 = NULL;
#define	ARGBEGIN	for((void)(argv0||(argv0=*argv)),argv++,argc--;\
			    argv[0] && argv[0][0]=='-' && argv[0][1];\
			    argc--, argv++) {\
				char *_args, *_argt;\
				char _argc;\
				_args = &argv[0][1];\
				if(_args[0]=='-') {\
					if(_args[1]==0){\
						argc--; argv++; break;\
					} else {\
						_args++;\
						_args[1]=0;\
					}\
				}\
				_argc = 0;\
				while(*_args && (_argc = *_args++))\
				switch(_argc)
#define	ARGEND		SET(_argt);USED(_argt);USED(_argc);USED(_args);}\
					USED(argv);USED(argc);
#define	ARGF()		(_argt=_args, _args="",\
				(*_argt? _argt: argv[1]? (argc--, *++argv): 0))
#define	EARGF(x)		(_argt=_args, _args="",\
				(*_argt? _argt: argv[1]? (argc--, *++argv): ((x), abort(), (char*)0)))

#define	ARGC()		_argc

#define	SET(x)	(x) = 0
#define	USED(x)	(void)(x)

int
getfields(char *str, char **args, int max, int mflag, char *set)
{
	char c;
	int intok, narg;

	if(max <= 0)
		return 0;

	narg = 0;
	args[narg] = str;
	if(!mflag)
		narg++;
	intok = 0;
	for(;; str++) {
		c = *str;
		if(c == 0)
			break;
		if(strchr(set, c)) {
			if(narg >= max)
				break;
			*str = 0;
			intok = 0;
			args[narg] = str + 1;
			if(!mflag)
				narg++;
		} else {
			if(!intok && mflag)
				narg++;
			intok = 1;
		}
	}
	return narg;
}
/* End Plan 9 stuff. */

void
usage(void)
{
	fprintf(stderr, "Usage: %s [-i input] [-o output] [-lNh] [column ...]\n", argv0);
	exit(1);
}

void
verror(char *fmt, va_list arg)
{
	char buf[1024];
	vsnprintf(buf, sizeof(buf), fmt, arg);
	if (argv0)
		fprintf(stderr, "%s: %s\n", argv0, buf);
	else
		fprintf(stderr, "%s\n", buf);
}

void
warn(char *fmt, ...)
{
	va_list arg;
	va_start(arg, fmt);
	verror(fmt, arg);
	va_end(arg);
}

void
fatal(char *fmt, ...)
{
	va_list arg;
	va_start(arg, fmt);
	verror(fmt, arg);
	va_end(arg);
	exit(1);
}

void *
emalloc(int size)
{
	void *p = malloc(size);
	if (p == NULL)
		fatal("malloc");
	return p;
}

void *
erealloc(void *p, int size)
{
	p = realloc(p, size);
	if (p == NULL)
		fatal("realloc");
	return p;
}

FILE *
efreopen(const char *path, const char *mode, FILE * stream)
{
	FILE *f = freopen(path, mode, stream);
	if (f == NULL)
		fatal("freopen %s: %s", path, strerror(errno));
	return f;
}

/* Clone of Bio's Brdstr */
char *
frdstr(FILE *f, char delim, int nulldelim)
{
	char *buf;
	int size=0, got=0;
	int c;

	size = 128;
	buf = malloc(size);

	while ((c = getc(f)) != EOF) {
		if (got >= size-1) {	/* size-1 so we can add a NUL */
			size *= 1.5;
			buf = erealloc(buf, size);
		}
		buf[got++] = c;
		if (c == delim)
			break;
	}
	if (got == 0) {
		free(buf);
		if (ferror(f))
			fatal("read error: %s", strerror(errno));
		return NULL;
	}
	if (nulldelim && buf[got-1] == delim)
		buf[got-1] = '\0';
	else
		buf[got] = '\0';

	return buf;
}

int
main(int argc, char **argv)
{
	char **f;
	char *l;
	int *fields;
	int ncol, nf;
	int line, i, j;
	int last = 0;
	int header = 1;

	ARGBEGIN{
	case 'i':
		efreopen(EARGF(usage()), "r", stdin);
		break;
	case 'o':
		efreopen(EARGF(usage()), "w", stdout);
		break;
	case 'l':
		last = 1;
		break;
	case 'h':
		execlp("grep", "grep", "-v", "^#", HELPFILE, NULL);
		fatal("Couldn't exec() grep: %s", strerror(errno));
		break;
	case 'n':
	case 'N':
		header=0;
		break;
	default:
		usage();
	}ARGEND

	if (argc == 0)
		exit(0);

	if (header) {
		for (i=0;argv[i] != NULL; i++)
			fprintf(stdout, "%s\001%s", i > 0 ? "\t": "", argv[i]);
		putc('\n', stdout);
	}

	fields = emalloc(argc * sizeof(int));

	l = frdstr(stdin, '\n', 1);
	if (l == NULL || l[0] != '\001')
		fatal("not a valid table");

	for (i=0,ncol=0; l[i] != '\0'; i++)
		if (l[i] == '\001')
			ncol++;

	/* extra column for detecting corrupt rows */
	f = emalloc((ncol+1) * sizeof(char *));

	nf = getfields(l, f, ncol+1, 0, "\t");
	if (nf != ncol)
		fatal("corrupt header line");
	
	for (i=0; i<argc; i++) {
		fields[i] = -1;
		for (j=0; j<ncol; j++) {
			if (strcmp(argv[i], f[j]+1) == 0) {	/* +1 skips \001 */
				fields[i] = j;
				if (!last)
					break;
			}
		}
	}

	free(l);

	for (line=2; (l = frdstr(stdin, '\n', 1)) != NULL; free(l), line++) {
		/* We look for an extra column in case the table is corrupt. */
		nf = getfields(l, f, ncol+1, 0, "\t");
		if (nf != ncol) {
			warn("wrong number of columns on line %d, skipping", line);
			continue;
		}
		for (i=0; i < argc; i++) {
			if (i > 0)
				putc('\t', stdout);
			if (fields[i] >= 0)
				fputs(f[fields[i]], stdout);
		}
		putc('\n', stdout);
	}

	return 0;
}
